/* -----------------------------------------------------------------------------
 * d.cxx
 *
 *     D wrapper module.
 *
 * Author(s) : Andy Friesen
 *
 * Copyright (C) 2004
 * See the file LICENSE for information on usage and redistribution.
 * ----------------------------------------------------------------------------- */

char cvsroot_d_cxx[] = "$Header: /cvsroot/SWIG/Source/Modules/d.cxx,v 1.50 2003/12/23 21:45:38 cheetah Exp $";

#include <limits.h> // for INT_MAX
#include <ctype.h>
#include "swigmod.h"
#ifndef MACSWIG
#   include "swigconfig.h"
#endif

#include <cassert>
#include <string>
#include <vector>
#include <map>
#include <iostream>

namespace {

    template <typename T>
    void destroy(T* t) {
        Delete(t);
    }

    template<>
    void destroy<Wrapper>(Wrapper* w) {
        DelWrapper(w);
    }

    struct ScopedPtrBase { };

    template <class T>
    struct ScopedPtr : ScopedPtrBase {
        ScopedPtr(T* t)  : _data(t) { }
        ScopedPtr()      : _data(0) { }

        ScopedPtr(ScopedPtr& rhs) {
            (*this) = rhs;
        }

        virtual ~ScopedPtr() { 
            destroy(_data);
        }

        T* operator ->() const { return  _data; }
        T* get()         const { return  _data; }
     
        ScopedPtr& operator = (T* t) {
            destroy(_data);
            _data = t;
            return *this;
        }

        ScopedPtr& operator = (ScopedPtr& rhs) {
            (*this) = rhs.get();
            rhs = 0;
            return *this;
        }

    private:
        T* _data;
        //ScopedPtr(ScopedPtr&) { } // disallow copying
    };

    typedef ScopedPtr<File> ScopedFile;
    typedef ScopedPtr<String> ScopedString;

    template <typename T>
    ScopedPtr<T> makeAutoPtr(T* t) {
        return ScopedPtr<T>(t);
    }
    struct DObject {
        DObject(DOH* d = 0, bool own = false)
            : _data(d)
            , _own(own)
        { }

        ~DObject() {
            if (_own) {
                destroy(_data);
            }
        }

        // Maybe make this private.
        DObject& operator = (DOH* rhs) {
            if (_own) {
                destroy(_data);
            }
            _data = rhs;
            _own = false; // what else do we do? :)
            return *this;
        }

        DOH* get() const {
            return _data;
        }

        void release() {
            _own = false;
        }

        bool empty() const {
            return get() == 0;
        }

        static bool check(DOH*) {
            return true;
        }
    private:
        DOH* _data;
        bool _own;
    };

    struct DHash : DObject {
        DHash(DOHHash* p = 0, bool own = false)
            : DObject(p, own)
        {}

        DOH* getattr(DOH* key) const {
            return Getattr(get(), key);
        }

        template <typename T>
        T getattr(DOH* key, T defaultValue) {
            if (has(key)) {
                return getattr(key);
            } else {
                return defaultValue;
            }
        }

        void setattr(DOH* key, DOH* value) {
            Setattr(get(), key, value);
        }

        bool has(DOH* key) {
            return getattr(key) != 0;
        }

        struct ValueProxy {
            friend struct DHash;

            operator DOH*() const {
                return _value;
            }

            ValueProxy& operator = (DOH* rhs) {
                _this->setattr(_key, rhs);
                return *this;
            }

            ValueProxy(const ValueProxy& rhs)
                : _this(rhs._this)
                , _key(rhs._key)
                , _value(rhs._value)
            { }

        private:
            ValueProxy(DHash* parent, DOH* key)
                : _this(parent)
                , _key(key)
                , _value(parent->getattr(key))
            { }

            DHash* _this;
            DOH* _key;
            DOH* _value;
        };

        DOH* operator[](DOH* key) const {
            return getattr(key);
        }

        ValueProxy operator[](DOH* key) {
            return ValueProxy(this, key);
        }
        DHash& operator = (const ValueProxy& rhs) {
            *this = rhs._value;
            return *this;
        }

        static bool check(DOH* d) {
            return DohIsMapping(d);
        }
    };

    struct DString : DObject {
        DString(DOH* p = 0, bool own = false)
            : DObject(p, own)
        { }

        bool contains(const DString& rhs) {
            if (empty()) {
                return rhs.empty();
            } else if (rhs.empty()) {
                return false;
            } else {
                // strstr doesn't like nulls.  I don't want to care.
                return Strstr(get(), rhs.get()) != 0;
            }
        }

        bool operator == (DOH* rhs) {
            return rhs == get();
        }

        bool operator == (const DString& rhs) {
            return *this == rhs.get();
        }

        static bool check(DOH* d) {
            return DohIsString(d);
        }
    };

    using std::ostream;
    using std::endl;
    using std::cerr;

    ostream& operator << (ostream& lhs, const DString& rhs) {
        lhs << DohData(rhs.get());
        return lhs;
    }
}

// Evil little macro for automatically destroying a local when its scope dies.
// magic: Assigning a temporary to a const& extends the life of the temporary
// to that of the reference.
#define AUTO_HIDDEN(x, y) x ## y
#define AUTO_HIDE2(x, y) AUTO_HIDDEN(x, y)
#define AUTO(x) ScopedPtrBase const& AUTO_HIDE2(auto_secret, __LINE__) = makeAutoPtr(x);

class D : public Language {
    /*
     * TODO: deal with nested types.
     */

    virtual void main(int argc, char** argv) {
        SWIG_library_directory("d");
        Preprocessor_define("SWIGD 1", 0);
        SWIG_config_file("d.swg");
        SWIG_typemap_lang("d");

        ScopedString cname = NewStringf("test.cpp");
        ScopedString dimp  = NewStringf("test_extern.d");
        ScopedString dname = NewStringf("test.d");

        c_source = NewFile(cname.get(), "w");
        d_import = NewFile(dimp .get(), "w");
        d_source = NewFile(dname.get(), "w");

#if 0
        c_init   = NewString("");
#endif

        // SWIG uses these names to insert code in various places.
        // Link them up with actual file handles.
        Swig_register_filebyname("runtime", c_source.get());
        Swig_register_filebyname("wrapper", d_source.get());

        // We don't actually have a C header: 
        // just put it in the main C++ source file for now.
        Swig_register_filebyname("header", c_source.get());

        allow_overloading();
    }

    virtual int top(Node* n) {
        Language::top(n);
        return SWIG_OK;
    }

    //
    virtual int enumDeclaration(Node* n) {
        DHash node(n);

        Printf(stderr, "enum %s\n", node["name"]);

        Printf(d_source.get(), "enum %s {\n", node["name"]);
        Language::enumDeclaration(n);
        Printf(d_source.get(), "}\n");
        return SWIG_OK;
    }

    virtual int enumvalueDeclaration(Node* n) {
        DHash node(n);

        Printf(d_source.get(), "    %s = %s,\n", node["name"], node["enumvalue"]);
        return SWIG_OK;
    }

    //
    virtual int globalfunctionHandler(Node* n) {
        DHash node(n);

        // Already handled
        if (node.has("funcType")) {
            return Language::globalfunctionHandler(node.get());
        }

        int result = SWIG_ERROR;
        node["funcType"] = "global";

        if (node.has("sym:overloaded") && node["storage"] == "static") {
            // overloads need proxies so overloading works correctly.
            // pass-through to generate proxies.
            Printf(stderr, "global overloaded function %s\n", node["name"]);

            result = Language::globalfunctionHandler(node.get());
        } else {
            String* s = node["name"];
            Printf(stderr, "Skipping non-overloaded global function %s: D needs no glue.\n", node["name"]);
            result = SWIG_NOWRAP;
        }

        node["funcType"] = 0;
        return result;
    }

    virtual int functionHandler(Node* n) {
        DHash node(n);

        if (node.has("sym:overloaded")) {
            // Overloads aren't jiggy with extern "C".  Need proxies.
            // add wart to name to make it unique.
            node["sym:name"] = NewStringf("%s%s", node["sym:name"], node["sym:overname"]);
        }
        return Language::functionHandler(node.get());
    }

    virtual int memberfunctionHandler(Node* n) {
        DHash node(n);

        Printf(stderr, "\tmethod: %s\n", node["name"]);

        node["funcType"] = "method";
        int result = Language::memberfunctionHandler(n);
        node["hasThis"] = 0;
        return result;
    }

    virtual int staticmemberfunctionHandler(Node* n) {
        DHash node(n);

        Printf(stderr, "\tstatic method %s\n", node["name"]);
        node["funcType"] = "staticmethod";
        dumpNode(n);
        // skip the default static method handler: it mangles names
        int result = Language::globalfunctionHandler(n);
        node["funcType"] = 0;
        dumpNode(n);
        return result;
    }

    virtual int callbackfunctionHandler(Node* n) {
        return NYI(n);
    }

    //
    virtual int globalvariableHandler(Node *n) {
        DHash node(n);

        Printf(stderr, "\tglobal attribute %s\n", node["name"]);
        return Language::globalvariableHandler(n);
    }

    virtual int membervariableHandler(Node *n) {
        DHash node(n);

        Printf(stderr, "\tattribute %s\n", node["name"]);

        // Set a flag so writeWrapper knows
        node["funcType"] = "attribute";
        int result = Language::membervariableHandler(n);
        node["funcType"] = 0;
        return result;
    }

    virtual int staticmembervariableHandler(Node *n) {
        DHash node(n);

        Printf(stderr, "\tstatic attribute %s\n", node["name"]);

        node["funcType"] = "staticattribute";
        int result = Language::staticmembervariableHandler(n);
        node["funcType"] = 0;
        return result;
    }

    //
    virtual int memberconstantHandler(Node* n) {
        DHash node(n);

        Printf(stderr, "\tconstant attribute %s\n", node["name"]);
        return Language::memberconstantHandler(n);
    }

    virtual int constructorHandler(Node* n) {
        DHash node(n);

        Printf(stderr, "\tctor %s\n", node["name"]);

        node["name"] = "this";
        node["funcType"] = "ctor";
        int result = Language::constructorHandler(n);
        node["funcType"] = 0;
        return result;
    }

    virtual int copyconstructorHandler(Node* n) {
        DHash node(n);

        Printf(stderr, "\tcopy ctor %s\n", node["name"]);                

        node["name"] = "this";
        node["funcType"] = "ctor";
        int result = Language::copyconstructorHandler(n);
        node["funcType"] = 0;
        return result;
    }

    virtual int destructorHandler(Node* n) {
        DHash node(n);

        Printf(stderr, "\tdtor\n");

        node["name"] = "~this";
        node["funcType"] = "dtor";
        int result = Language::destructorHandler(n);
        node["funcType"] = 0;
        return result;
    }

    // FIXME: Too complex.  Simplify.
    virtual int classHandler(Node* n) {
        DHash node(n);

        Printf(stderr, "\nclass %s:\n", node["name"]);
        
        String* className = node["name"];
        String* baseClause = NewString("");
        AUTO(baseClause);

        List* baseList = node["bases"];
        if (baseList != 0) {
            Iterator iter = First(baseList);
            if (iter.item) {
                Printf(baseClause, " : %s", Getattr(iter.item, "name"));
            }
            if (Next(iter).item != 0) {
                // As good a warning as any.
                Swig_warning(WARN_JAVA_MULTIPLE_INHERITANCE, input_file, line_number,
                    "Warning for %s proxy: Base %s ignored. Multiple inheritance "
                    "is not supported in D.\n", className, Getattr(iter.item,"name")
                );
            }
        }

        // FIXME: show base(s)
        Printf(d_source.get(), "class %s%s {\n", className, baseClause);
        // impl pointer, protected constructor
        Printv(d_source.get(), tab4, "private void* swig_impl;\n", NIL);
        if (baseList == 0) {
            Printv(d_source.get(), tab4, "private bool  swig_own_ptr;\n", NIL);
            Printv(d_source.get(), 
                tab4, "protected this(void* ptr, bool own) {\n",
                tab4, "    swig_impl = ptr;\n",
                tab4, "    swig_own_ptr = own;\n",
                tab4, "}\n",
                NIL
            );
        } else {
            Printv(d_source.get(),
                tab4, "protected this(void* ptr, bool own) {\n",
                tab4, "    super(ptr, own);\n",
                tab4, "    swig_impl = ptr;\n",
                tab4, "}\n",
                NIL
            );
        }

        Language::classHandler(n);

        Printf(d_source.get(), "}\n\n");

        return SWIG_OK;
    }

    //
    virtual int constantWrapper(Node* n) {
        DHash node(n);

        String* varType = node["type"];
        String* varName = node["name"];
        String* varValue = node["value"];

        int valueType = SwigType_type(varType);
        if (valueType == T_STRING || valueType == T_CHAR) {
            // put the quotes back (SWIG strips them for some reason I cannot fathom)
            char quote = valueType == T_STRING ? '\"' : '\'';
            String* oldValue = varValue;

            varValue = NewString("");

            Printf(varValue, "%c%s%c", quote, oldValue, quote);
            node["value"] = varValue;
        }

        Printf(d_source.get(), "const %s %s = %s;\n", varType, varName, varValue);

        return SWIG_OK;
    }

    virtual int functionWrapper(Node* node) {
        int result;

        // C glue
        result = writeExternCFunc(node);
        if (result != SWIG_OK) {
            return result;
        }

        // D glue
        result = writeExternDFunc(node);
        if (result != SWIG_OK) {
            return result;
        }

        // wrapper interface
        result = writeWrapper(node);
        return result;
    }

    int writeExternCFunc(Node* n) {
        DHash node(n);

        String*    name  = node["sym:name"];
        SwigType*  type  = node["type"];
        ParmList*  parms = node["parms"];

        String* returnType = SwigType_str(type, 0);
        AUTO(returnType);

        Wrapper* wrapper = NewWrapper();
        AUTO(wrapper);

        emit_attach_parmmaps(parms, wrapper);
        Swig_typemap_attach_parms("CtoExtern", parms, wrapper);

        Printf(wrapper->def, "DllExport %s CDecl %s(%s) {", returnType, name, ParmList_str(parms));

        emit_args(type, parms, wrapper);

        emit_action(node.get(), wrapper);

        if (SwigType_type(type) != T_VOID) {
            Printf(wrapper->code, "return result;");
        }

        Printf(wrapper->code, "}");
        Wrapper_print(wrapper, c_source.get());

        writeExternDFunc(node.get());

        return SWIG_OK;
    }

    int writeExternDFunc(Node* n) {
        DHash node(n);

        String*    name  = node["sym:name"];
        SwigType*  type  = node["type"];
        ParmList*  parms = node["parms"];

        String*    argList = NewString("");
        AUTO(argList);

        String* returnType = SwigType_str(type, 0);
        AUTO(returnType);

        Swig_typemap_attach_parms("CtoD", parms, 0);

        String* mappedReturnType = Swig_typemap_lookup_new("CtoD", node.get(), "", 0);

        if (mappedReturnType == 0) {
            mappedReturnType = Copy(returnType);
        }

        // Construct argument list: transform each argument type
        // and make up a name.

        int parmCount = emit_num_arguments(parms);
        DHash parm = parms;
        for (int i = 0; i < parmCount; i++) {
            String* mappedType = parm["tmap:CtoD"];

            String* oldType = SwigType_str(parm["type"], 0);

            if (mappedType != 0) {
                Printf(stderr, "Can't convert argument type %s in function %s\n", oldType, name);
                mappedType = oldType;
            }
            
            Printv(argList, "%s arg%i", mappedType, i + 1, NIL);

            if (i < parmCount - 1) {
                Printf(argList, ", ");
            }

            parm = parm["nextSibling"];
        }

        Printf(d_import.get(), "extern (C) %s %s(%s);\n", mappedReturnType, name, argList);

        return SWIG_OK;
    }

    int writeWrapper(Node* n) {
        DHash node(n);

        // Now the D wrapper
        if (!node.has("funcType")) {
            Printf(stderr, "Internal error on function %s in D::writeWrapper.  Maybe.\n", node["name"]);
            return SWIG_ERROR;
        }

        DString funcType = node["funcType"];
        bool isAttribute = funcType =="attribute";
        bool isCtor = funcType == "ctor";
        bool isDtor = funcType == "dtor";
        bool isMethod = funcType == "method";
        bool isStaticMethod = funcType == "staticmethod";
        bool isGlobalMethod = true; // :P

        int result = SWIG_ERROR;
        if (isCtor) {
            result = writeWrapperCtor(node.get());

        } else if (isDtor) {
            result = writeWrapperDtor(node.get());

        } else if (isAttribute) {
            Printf(stderr, "Skipping attribute %s\n", node["name"]);
            result = SWIG_OK;

        } else if (isMethod || isStaticMethod || isGlobalMethod) {
            // Hack: need to restore the name on static methods
            if (isStaticMethod) {
                node["sym:name"] = node["name"];
            }
            result = writeWrapperFunc(node.get());

        } else {
            Printf(stderr, "Skipping %s %s: D needs no wrapper for pure C things.\n", 
                node["nodeType"], node["name"]
            );

            result = SWIG_OK;
        }

        return result;
    }

    int writeWrapperFunc(Node* n) {
        DHash node(n);

        String*   name  = node["name"];        // pretty name
        String*   wrapName = node["sym:name"]; // mangled name
        SwigType* returnType = node["type"];        // return type
        ParmList* parms = node["parms"];

        const bool   isStatic = DString(node["storage"]).contains("static");
        const bool    hasThis = !DString(node["funcType"]).contains("static");
        const bool returnsVoid = SwigType_type(returnType) != T_VOID;

        Wrapper* wrapper = NewWrapper();
        AUTO(wrapper);

        Swig_typemap_attach_parms("DtoC", parms, wrapper);

        String* argList = NewString("");
        AUTO(argList);

        DHash parm = parms;
        int parmCount = emit_num_arguments(parms);

        if (hasThis && !parm.empty()) {
            // skip the first argument if we have a 'this' pointer
            parm = parm["nextSibling"];
            parmCount--;
        }

        for (int i = 0; i < parmCount; i++) {
            String* mappedType = parm["tmap:CtoD"];

            String* oldType = SwigType_str(parm["type"], 0);

            if (mappedType != 0) {
                Printv(argList, mappedType, NIL);
            } else {
                Printf(stderr, "Can't convert argument type %s in function %s\n", oldType, wrapName);
                Printv(argList, oldType, NIL);
            }

            Printf(argList, " arg%i", i + 1);

            if (i < parmCount - 1) {
                Printv(argList, ", ", NIL);
            }

            parm = parm["nextSibling"];
        }

        if (isStatic) {
            Printf(wrapper->def, "static ");
        }

        Printf(wrapper->def, "%s %s(%s) {", SwigType_str(returnType, 0), name, argList);

        // Actually build the method body
        if (hasThis) {
            Printf(wrapper->code, "void* arg1 = swig_impl;\n");
        }

        if (returnsVoid) {
            Printv(wrapper->code, "return ", NIL);
        }

        Printv(wrapper->code, Swig_cfunction_call(wrapName, parms), ";", NIL);

        Printf(wrapper->code, "}");
        // --

        Wrapper_print(wrapper, d_source.get());

        return SWIG_OK;
    }

    int writeWrapperCtor(Node* n) {
        DHash node(n);

        ParmList* parms = node["parms"];
        String* name = node["name"];
        AUTO(name);

        Wrapper* wrapper = NewWrapper();
        AUTO(wrapper);

        Swig_typemap_attach_parms("DtoC", parms, wrapper);

        String* argList = NewString("");
        AUTO(argList);

        int parmCount = emit_num_arguments(parms);
        DHash iter = parms;
        for (int i = 0; i < parmCount; i++) {
            SwigType* parmType = iter["type"];
            String* parmName = iter["name"];

            String* mappedType = iter["tmap:DtoC"];
            if (mappedType != 0) {
                Printv(argList, mappedType, NIL);
            } else {
                String* oldType = SwigType_str(iter["type"], 0);
                Printf(stderr, "Can't convert argument type %s in function %s\n", oldType, parmName);
                Printv(argList, oldType, NIL);
            }

            if (i < parmCount - 1) {
                Printv(argList, ", ", NIL);
            }

            iter = iter["nextSibling"];
        }

        Printf(d_source.get(), "    // ctor\n");

        Printf(stderr, "CTOR\t%s\n", argList);

        return SWIG_OK;
    }

    int writeWrapperDtor(Node* n) {
        DHash node(n);

        // Base class defines protected swig_destroy() and ~this()
        // Subclasses just override swig_destroy()
        String* dtorName = node["sym:name"];
        if (!node.has("bases")) {
            Printf(d_source.get(),
                "    protected void swig_destroy() {\n"
                "        if (swig_own_ptr && swig_impl !== null) {\n"
                "            %s();\n"
                "            swig_impl = null;\n"
                "        }\n"
                "    }\n"
                "\n"
                "    ~this() {\n"
                "        swig_destroy();\n"
                "    }\n"
                "\n",
                dtorName
            );
        } else {
            // Suppress destructor on subclasses
            Printf(d_source.get(), 
                "    protected void swig_destroy() {\n"
                "        if (swig_own_ptr && swig_impl !== null) {\n"
                "            %s();\n"
                "            swig_impl = null;\n"
                "        }\n"
                "    }\n"
                "\n",
                dtorName
            );
        }
        return SWIG_OK;
    }

    void dumpNode(const DHash& node) {
        dumpNode(node.get());
    }

    void dumpNode(Node* node) {
            for (Iterator iter = First(node); iter.item; iter = Next(iter)) {
            if (DohIsString(iter.item)) {
                Printf(stderr, "'%s'\t'%s'\n", iter.key, iter.item);
            } else {
                Printf(stderr, "'%s'\t<%s>\n", iter.key, Getattr(iter.item, "nodeType"));
            }
        }
        Printf(stderr, "\n\n\n");
    }

    int NYI(Node* n) {
        DHash node(n);
        Printf(stderr, "Not yet implemented: %s %s\n", node["nodeType"], node["name"]);
        dumpNode(node.get());
        return SWIG_ERROR;
    }

    // attributes:

    /// Contains the implementation of the wrapped C API
    ScopedFile c_source;

    /// D source file that declares extern(C) functions for 
    /// the contents of c_source (above)
    ScopedFile d_import;

    /// Wraps the C API in a delicious, candy coated interface.
    ScopedFile d_source;

    /// Not used. (not useful?  Provide anyway for the sake of compatibility with other backends?)
#if 0
    ScopedString c_init;   // init code. (also goes into c_source, I think)
#endif

};   /* class D */

extern "C" Language *
swig_d(void) {
  return new D();
}

#if 0
const char *D::usage = reinterpret_cast<char*>("\
D Options (available with -d)\n\
     -package <name> - set name of the package to <name>\n\
\n");
#endif
